package buffer;
import java.util.Arrays;
import java.util.Comparator;
/**
 * A SortedBuffer is a buffer of limited size that keeps n comparable elements 
 * sorted in ascending order. The elements are stored consecutively, 
 * with the first element in position 0.
 * 
 * @author Uno Holmer
 * @version 2016-12-05
 * @Invariant size >= 0 && size < buffer.length
 */
    public class SortedBuffer<T> implements Buffer<T> {
    private int size;         // the number of stored elements
    private T[] buffer;       // the elements
    private Comparator<? super T> comp;
    private boolean useNaturalOrderComparator;

    /**
     * Creates an empty buffer of the given capacity.
     * @param comp comparator for the elements
     * @param capacity the maximum number of elements that can be stored in this buffer.
     * @Pre capacity > 0
     * @Post invariant()
     */
    @SuppressWarnings("unchecked")
    public SortedBuffer(Comparator<? super T> comp,int capacity) {
        assert capacity > 0;
        buffer = (T[])new Object[capacity];
        size = 0;
        if ( comp != null ) {
            this.comp = comp;
            useNaturalOrderComparator = false;
        } else {
            this.comp = (Comparator<? super T>)Comparator.naturalOrder();
            useNaturalOrderComparator = true;
        }
        assert invariant();
    }
   
    /**
     * Creates an empty buffer of the given capacity.
     * @param capacity the maximum number of elements that can be stored in this buffer.
     * @Pre capacity > 0
     * @Post invariant()
     */
    public SortedBuffer(int capacity) {
        this(null,capacity);
    }
    
    public boolean isEmpty() {
        return size == 0;
    }
    
    public boolean isFull() {
        return size == buffer.length;
    }

    public int getSize() {
        return size;
    }

    public int getCapacity() {
        return buffer.length;
    }
    
    /**
     * Find the i'th element of this buffer.
     * The call succeeds if the position of the specified element
     * is within the range of the stored elements, otherwise it fails.
     * The operation does not alter the contents of the buffer.
     * 
     * @param      i the position of the wanted element.
     * @return     the element at the specified position.
     * @throw      IndexOutOfBoundsException
     */
    public T get(int i) throws IndexOutOfBoundsException {
        if ( i < 0 || i >= size )
            throw new IndexOutOfBoundsException();
        else
            return buffer[i];
    }

    /**
     * Insert an object into this buffer.
     * If the insertion does not violate 
     * the buffer's capacity it succeeds, otherwise it fails. 
     * If size() returns n before a successful insertion, it will
     * return n+1 after that insertion. 
     * The new element is inserted in a position that is consistent
     * with the ordering among the n+1 elements.
     *
     * @param  o   the object to be inserted
     * @throw  IllegalStateException if the buffer is full.
     * @throw  ClassCastException if x is not comparable. 
     * @Pre !isFull()
     * @Post invariant()
     */
    public void insert(T o) throws IllegalStateException,ClassCastException {
        assert invariant();
        checkIfComparable(o);
        checkForSpace();
        doTheInsertion(o);
        assert invariant();
    }
    
    private void checkIfComparable(T x) throws ClassCastException {
        if ( useNaturalOrderComparator && ! (x instanceof java.lang.Comparable) )
            throw new ClassCastException(x.getClass().getName() + 
                                         " cannot be cast to java.lang.Comparable");
    }
    
    private void checkForSpace() throws IllegalStateException {
        if ( size >= buffer.length )
            throw new IllegalStateException();
    }
    
    private void doTheInsertion(T x) {
        int i = size;    
        while ( i > 0 && comp.compare(buffer[i-1],x) > 0 ) {
            buffer[i] = buffer[i-1];
            i--;
        }
        buffer[i] = x;
        size++;
    }
    
    /**
     * Remove an element from this buffer.
     * The removal succeeds if the position of the specified element
     * is within the range of the stored elements, otherwise it fails.
     * After a successful removal, the remaining elements are packed in consecutive
     * ascending order. If size returns n (n > 0) before the removal, 
     * it will return n-1 after a successful removal.
     * The stored elements are left untouched in the case of failure.
     *
     * @param  i   the position of the element to be removed.
     * @throw IndexOutOfBoundsException if the argument is 
     *        not within the limits of this buffer.
     * @Pre i >= 0 && i < getSize()
     * @Post invariant()
     */
    public void remove(int i) throws IndexOutOfBoundsException {
        assert invariant();
        if ( i < 0 || i >= size )
            throw new IndexOutOfBoundsException();
            
        for ( int j = i; j < size - 1; j++ )
            buffer[j] = buffer[j+1];
        size--;
        assert invariant();
    }
    
    /**
     * The class invariant
     */
    public boolean invariant() {
        return size >= 0 && size <= buffer.length; // && isOrdered(); see comment below
    }
 
//	  We will not use this method for now. Instead, we will try to black box test
//    the ordering property by solely using the other public methods.
//    /**
//     * Determine if the elements in this buffer are sorted in ascending order.
//     *
//     * @return     true if all elements are ordered, false otherwise.
//     */
//    public boolean isOrdered(){
//        for ( int i = 0; i < size - 1; i++ )
//            if ( comp.compare(buffer[i],buffer[i+1]) > 0 )
//                return false;
//        return true;
//    }
    
    public String toString() {
        StringBuilder buf = new StringBuilder();
        buf.append('[');
        for ( int i = 0; i < size; i++ ) {
            buf.append("" + buffer[i]);
            if ( i < size-1 )
                buf.append(',');
        }
        buf.append(']');
        return buf.toString();
    }
}



















