package buffer;
import org.junit.Before;
import org.junit.After;
import org.junit.Test;
import static org.junit.Assert.*;
import java.util.*;
/**
 * The test class SortedBufferTest.
 *
 * @author  Uno Holmer
 * @version 2016-11-29
 */
public class SortedBufferTest {
	// Test fixture
    private SortedBuffer<Integer> emptyBuffer;
    private SortedBuffer<Integer> fullBuffer;
    private SortedBuffer<Integer> buffer3;
    
    private final int[][] permutations;	// Test cases

    public SortedBufferTest() {
        permutations = 
            new int[][] {
                    {1,1,1},	// All equal
                    {1,1,2},	// Two equal
                    {1,2,1},
                    {2,1,1},
                    {2,2,1},
                    {2,1,2},
                    {1,2,2},
                    {1,2,3},	// None equal 
                    {1,3,2},
                    {2,1,3},
                    {2,3,1},
                    {3,1,2},
                    {3,2,1}
                };
    }

    /**
     * Sets up the test fixture.
     *
     * Called before every test case method.
     */
    @Before
    public void setUp() {
        emptyBuffer = new SortedBuffer<Integer>(1);
        fullBuffer = new SortedBuffer<Integer>(1);
        fullBuffer.insert(123);
        buffer3 = new SortedBuffer<Integer>(3);
    }

    /**
     * Tears down the test fixture.
     *
     * Called after every test case method.
     */
    @After
    public void tearDown() {}
        
    // Simple cases
    
    @Test
    public void testIsEmpty() {
        assertTrue(emptyBuffer.isEmpty());
    }
    
    @Test
    public void testIsFull() {
        assertTrue(fullBuffer.isFull());
    }

    @Test
    public void testGetSize() {
        assertEquals(0,emptyBuffer.getSize());
        assertEquals(1,fullBuffer.getSize());
    }    
    
    @Test
    public void testGetCapacity() {
        assertEquals(1,emptyBuffer.getCapacity());
        assertEquals(1,fullBuffer.getCapacity());
    }
    
//    @Test(expected = IndexOutOfBoundsException.class)
//    public void testGetOutOfBoundsLeft() {
//        fullBuffer.get(-1);
//    }
//    
//    @Test(expected = IndexOutOfBoundsException.class)
//    public void testGetOutOfBoundsRight() {
//        fullBuffer.get(1);
//    } 
    
    @Test(expected = IndexOutOfBoundsException.class)
    public void testRemoveEmpty() {
        emptyBuffer.remove(0);
    }
    
    @Test(expected = IndexOutOfBoundsException.class)
    public void testRemoveOutOfBoundsLeft() {
        fullBuffer.remove(-1);
    }
    
    @Test(expected = IndexOutOfBoundsException.class)
    public void testRemoveOutOfBoundsRight() {
        fullBuffer.remove(1);
    }    
    
    @Test(expected = IllegalStateException.class)
    public void testInsertFull() {
        fullBuffer.insert(456);
    }
    
    @Test
    public void testIncreasingSize() {
        assertEquals(3,buffer3.getCapacity());
        assertEquals(0,buffer3.getSize());
        for ( int i = 0; i < 3; i++ ) {
            int n = buffer3.getSize();
            buffer3.insert(123);
            assertEquals(n+1,buffer3.getSize());
        }
    }
    
    @Test
    public void testDecreasingSize() {
        assertEquals(3,buffer3.getCapacity());
        buffer3.insert(123);
        buffer3.insert(123);
        buffer3.insert(123);
        assertEquals(3,buffer3.getSize());
        for ( int i = 0; i < 3; i++ ) {
            int n = buffer3.getSize();
            buffer3.remove(0);
            assertEquals(n-1,buffer3.getSize());
        }
    }
   
    @Test
    public void testInsert() {
        for ( int[] permutation : permutations )
            testInsert(permutation);
    }
    
    private void testInsert(int[] permutation) {
        checkMatch(insertIntoBuffer(permutation),getSortedCopy(permutation),permutation);
    }
    
    private void checkMatch(SortedBuffer<Integer> buffer,int[] sortedArr,int[] originalArr) {
        assertEquals(buffer.getSize(),sortedArr.length);
        for ( int i = 0; i < sortedArr.length; i++ ) {
            assertEquals(errorMessage(i,buffer,sortedArr,originalArr),
                         (int)sortedArr[i],(int)buffer.get(i));
        }
    }
        
    private String errorMessage(int index,SortedBuffer<Integer> buffer,int[] sortedArr,int[] originalArr) {
        return "Expected: " + Arrays.toString(sortedArr) + 
               " but found " + buffer.toString() + 
               " in buffer, mismatch at index " + index + "\n" + 
               "Test case: " + Arrays.toString(originalArr) + "\n";
    }

    private SortedBuffer<Integer> insertIntoBuffer(int[] arr) {
        SortedBuffer<Integer> buffer = 
            new SortedBuffer<Integer>(arr.length);
        for ( int x : arr )
            buffer.insert(x);
        return buffer;
    }
    
    private int[] getSortedCopy(int arr[]) {
        int[] sortedCopy = (int[])arr.clone();
        Arrays.sort(sortedCopy);
        return sortedCopy;
    }
    
    @Test
    public void testRemove() {
        for ( int[] permutation : permutations )
            testRemove(permutation);
    }
    
    private void testRemove(int[] permutation) {
        for ( int i = 0; i < permutation.length; i++ ) {
            checkSizeAndCapacityAfterRemoval(i,permutation);
            checkMatchAfterRemoval(i,permutation);
        }
    } 
    
    private void checkSizeAndCapacityAfterRemoval(int indexOfElementToBeRemoved,int[] originalArr) {
         SortedBuffer<Integer> buffer = insertIntoBuffer(originalArr);
         int size = buffer.getSize();
         int capacity = buffer.getCapacity();
         buffer.remove(indexOfElementToBeRemoved);
         assertEquals(size - 1,buffer.getSize());
         assertEquals(capacity,buffer.getCapacity());
    }
    
    private void checkMatchAfterRemoval(int indexOfElementToBeRemoved,int[] originalArr) {
        SortedBuffer<Integer> buffer = insertIntoBuffer(originalArr);
        int[] sortedArr = skipElement(indexOfElementToBeRemoved,getSortedCopy(originalArr));
        buffer.remove(indexOfElementToBeRemoved);
        for ( int i = 0; i < sortedArr.length; i++ ) {
            assertEquals(errorMessage(i,buffer,sortedArr,originalArr),
                         (int)sortedArr[i],(int)buffer.get(i));
        }
    }
    
    private int[] skipElement(int indexOfElementToBeSkipped,int[] arr) {
        assert(indexOfElementToBeSkipped >= 0 && indexOfElementToBeSkipped < arr.length);
        int[] result = new int[arr.length-1];
        int i = 0;
        for (int j = 0; j < arr.length; j++)
            if ( j != indexOfElementToBeSkipped )
                result[i++] = arr[j];
        return result;
    }
}

































